import logging
from typing import *

from qtpy.QtCore import Slot
from qtpy.QtGui import QAction
from qtpy.QtWidgets import QMenu

from qcanvas import icons
from qcanvas.util import settings

_logger = logging.getLogger(__name__)


class _EnableAutoDownloadOption(QAction):
    def __init__(self, parent: Optional[QMenu] = None):
        super().__init__("Enable", parent)
        self.setCheckable(True)
        self.setChecked(settings.client.download_new_resources)
        self.triggered.connect(self._triggered)

    @Slot()
    def _triggered(self) -> None:
        settings.client.download_new_resources = self.isChecked()


class _EnableVideoDownloadOption(QAction):
    def __init__(self, parent: Optional[QMenu] = None):
        super().__init__("Include videos (slow)", parent)
        self.setCheckable(True)
        self.setChecked(settings.client.download_new_videos)
        self.triggered.connect(self._triggered)

    @Slot()
    def _triggered(self) -> None:
        settings.client.download_new_videos = self.isChecked()


class AutoDownloadResourcesMenu(QMenu):
    def __init__(self, parent: Optional[QMenu] = None):
        super().__init__("Auto download resources", parent)
        self.addAction(_EnableAutoDownloadOption(self))
        self.addAction(_EnableVideoDownloadOption(self))
        self.setIcon(icons.options.auto_download)



import asyncio
import logging
import sys

import qtpy
from qasync import QEventLoop
from qtpy.QtGui import QGuiApplication
from qtpy.QtWidgets import QApplication

import qcanvas.backend_connectors.qcanvas_task_master as task_master
from qcanvas.ui.main_ui.qcanvas_window import QCanvasWindow
from qcanvas.ui.setup import SetupDialog, setup_checker
from qcanvas.util import runtime, settings, themes

main_window = None
setup_window = None

_logger = logging.getLogger(__name__)


def _show_main():
    global main_window
    main_window = QCanvasWindow()
    main_window.show()


def _show_qt_api_name():
    print(f"Using Qt bindings from {qtpy.API_NAME}")
    _logger.info("Using Qt bindings from %s", qtpy.API_NAME)


def launch():
    _show_qt_api_name()

    app = QApplication(sys.argv)

    if runtime.is_running_as_flatpak:
        QGuiApplication.setDesktopFileName("io.github.qcanvas.QCanvasApp")

    app.setApplicationName("QCanvas")

    task_master.register()
    themes.apply(settings.ui.theme)

    event_loop = QEventLoop(app)
    asyncio.set_event_loop(event_loop)

    app_close_event = asyncio.Event()
    app.aboutToQuit.connect(app_close_event.set)

    if setup_checker.needs_setup():
        setup_window = SetupDialog()
        setup_window.show()
        setup_window.closed.connect(_show_main)
    else:
        _show_main()

    with event_loop:
        event_loop.run_until_complete(app_close_event.wait())



import logging
from abc import ABCMeta

from qcanvas_backend.task_master import TaskID, TaskMaster, set_global_task_master
from qtpy.QtCore import QObject, Signal

_logger = logging.getLogger(__name__)


class _Meta(type(QObject), ABCMeta): ...


class _TaskMaster(TaskMaster, QObject, metaclass=_Meta):
    task_progress = Signal(TaskID, int, int)
    task_failed = Signal(TaskID, object)

    def report_failed(self, _task_id: TaskID, context: object) -> None:
        self.task_failed.emit(_task_id, context)

    def report_progress(self, _task_id: TaskID, current: int, total: int) -> None:
        self.task_progress.emit(_task_id, current, total)


task_master = _TaskMaster()


def register():
    set_global_task_master(task_master)



# nuitka-project: --enable-plugin=pyside6
# nuitka-project: --windows-icon-from-ico=./deploy/windows/qcanvas.ico
# nuitka-project: --windows-console-mode=attach
# nuitka-project: --linux-icon=deploy/appimage/qcanvas.svg
# nuitka-project: --standalone
# nuitka-project: --include-module=yt_dlp

# Anti-bloat

##########################################################
# nuitka-project: --include-package=aiosqlite
# nuitka-project: --nofollow-import-to=aiosqlite.tests.*
##########################################################
# nuitka-project: --nofollow-import-to=rich
##########################################################
# nuitka-project: --nofollow-import-to=sqlalchemy.dialects.mssql
# nuitka-project: --nofollow-import-to=sqlalchemy.dialects.postgresql
# nuitka-project: --nofollow-import-to=sqlalchemy.dialects.oracle
# nuitka-project: --nofollow-import-to=sqlalchemy.dialects.mysql
##########################################################
# nuitka-project: --nofollow-import-to=rich
##########################################################
# nuitka-project: --nofollow-import-to=yt_dlp.extractor.lazy_extractors

import logging
from logging import INFO, WARNING

import qcanvas.app_start
from qcanvas.util import logs, paths

paths.data_storage().mkdir(parents=True, exist_ok=True)

logging.basicConfig(
    filemode="w",
    filename=paths.data_storage() / "debug.log",
    level="WARN",
)

logs.set_levels(
    {
        "qcanvas": INFO,
        "qcanvas.ui": WARNING,
        "qcanvas_backend": INFO,
        "qcanvas.ui.main_ui.status_bar_progress_display": INFO,
    }
)


def main():
    qcanvas.app_start.launch()


if __name__ == "__main__":
    main()



import logging

import qcanvas_backend.database.types as db
from qcanvas_backend.net.resources.download.resource_manager import ResourceManager
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qtpy.QtWidgets import *

from qcanvas.ui.course_viewer.tabs.assignment_tab.assignment_tree import AssignmentTree
from qcanvas.ui.course_viewer.tabs.content_tab import ContentTab
from qcanvas.ui.course_viewer.tabs.util import date_strftime_format
from qcanvas.util.basic_fonts import bold_label
from qcanvas.util.layouts import grid_layout

_logger = logging.getLogger(__name__)


class AssignmentTab(ContentTab):

    def __init__(
        self,
        *,
        course: db.Course,
        sync_receipt: SyncReceipt,
        downloader: ResourceManager,
    ):
        super().__init__(
            explorer=AssignmentTree.create_from_receipt(
                course, sync_receipt=sync_receipt
            ),
            title_placeholder_text="No assignment selected",
            downloader=downloader,
        )

        self._due_date_label = QLabel("")
        self._score_label = QLabel("")

        self.enable_info_grid()

    def setup_info_grid(self) -> QGridLayout:
        grid = grid_layout(
            [
                [
                    bold_label("Due:"),
                    self._due_date_label,
                ],
                [
                    bold_label("Score:"),
                    self._score_label,
                ],
            ]
        )

        grid.setColumnStretch(0, 0)
        grid.setColumnStretch(1, 1)

        return grid

    def update_info_grid(self, assignment: db.Assignment) -> None:
        if assignment.due_date is not None:
            due_text = assignment.due_date.strftime(date_strftime_format)
        else:
            due_text = "?"

        self._due_date_label.setText(due_text)

        self._score_label.setText(
            f"{assignment.mark or '?'}/{assignment.max_mark_possible or '?'}"
        )



from typing import Protocol

import qcanvas_backend.database.types as db
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt

date_strftime_format = "%A, %Y-%m-%d, %H:%M:%S"


# todo what the hell is the point of this?
class SupportsReload(Protocol):
    def reload(self, course: db.Course, *, sync_receipt: SyncReceipt) -> None: ...



import logging
from typing import *

import qcanvas_backend.database.types as db
from qcanvas_backend.net.resources.download.resource_manager import ResourceManager
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qtpy.QtCore import Slot
from qtpy.QtWidgets import *

from qcanvas.ui.course_viewer.content_tree import ContentTree
from qcanvas.ui.course_viewer.tabs.resource_rich_browser import ResourceRichBrowser
from qcanvas.util.basic_fonts import bold_font
from qcanvas.util.ui_tools import make_truncatable

_logger = logging.getLogger(__name__)

T = TypeVar("T", bound=Type["ContentTab"])


class ContentTab(QWidget):
    @classmethod
    def create_from_receipt(
        cls: T,
        *,
        course: db.Course,
        sync_receipt: SyncReceipt,
        downloader: ResourceManager,
    ) -> Type[T]:
        return cls(course=course, sync_receipt=sync_receipt, downloader=downloader)

    def __init__(
        self,
        *,
        explorer: ContentTree[db.Course],
        title_placeholder_text: str,
        downloader: ResourceManager,
    ):
        super().__init__()
        self._content_vbox = QVBoxLayout()
        self._placeholder_text = title_placeholder_text
        self._title_label = self._create_title_label()
        self._info_grid: Optional[QWidget] = None
        self._viewer = ResourceRichBrowser(downloader=downloader)
        self._explorer = explorer

        self._setup_layout()
        self._explorer.item_selected.connect(self._item_selected)

    def enable_info_grid(self) -> None:
        # Info grid needs to be a widget, so it can be hidden/shown
        grid_layout = self.setup_info_grid()

        grid_widget = QWidget()
        grid_widget.setLayout(grid_layout)
        grid_widget.hide()

        self._info_grid = grid_widget
        self._content_vbox.insertWidget(1, grid_widget)

    def _create_title_label(self) -> QLabel:
        title_label = QLabel(self._placeholder_text)
        title_label.setFont(bold_font)
        make_truncatable(title_label)
        return title_label

    def setup_info_grid(self) -> QGridLayout:
        """
        Override this if you need an info grid
        """
        raise NotImplementedError()

    def _setup_layout(self) -> None:
        parent_layout = QHBoxLayout()
        parent_layout.addWidget(self._explorer)

        self._content_vbox.addWidget(self._title_label)
        self._content_vbox.addWidget(self._viewer)

        parent_layout.addLayout(self._content_vbox)

        self.setLayout(parent_layout)

    def reload(self, course: db.Course, *, sync_receipt: SyncReceipt) -> None:
        self._explorer.reload(course, sync_receipt=sync_receipt)

    @Slot(object)
    def _item_selected(self, item: object) -> None:
        if isinstance(item, db.CourseContentItem):
            _logger.debug("Show %s", item.name)
            self._show_content(item)
        else:
            self._show_blank()

    def _show_content(self, item: db.CourseContentItem) -> None:
        self._title_label.setText(item.name)
        self._viewer.show_content(item)

        if self._info_grid is not None:
            self._info_grid.show()
            self.update_info_grid(item)

    def update_info_grid(self, content: db.CourseContentItem) -> None:
        raise NotImplementedError()

    def _show_blank(self) -> None:
        self._title_label.setText(self._placeholder_text)
        self._viewer.show_blank(completely_blank=True)

        if self._info_grid is not None:
            self._info_grid.hide()



import logging
from abc import abstractmethod
from typing import *

import qcanvas_backend.database.types as db
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qtpy.QtCore import QItemSelection, Signal, Slot
from qtpy.QtWidgets import *

from qcanvas.ui.course_viewer.tree_widget_data_item import AnyTreeDataItem
from qcanvas.ui.memory_tree import MemoryTreeWidget
from qcanvas.util.basic_fonts import bold_font, normal_font

_logger = logging.getLogger(__name__)

T = TypeVar("T")
U = TypeVar("U", bound=Type["ContentTree"])


class ContentTree(MemoryTreeWidget, Generic[T]):
    item_selected = Signal(object)

    @classmethod
    def create_from_receipt(
        cls: U, course: db.Course, *, sync_receipt: SyncReceipt
    ) -> Type[U]:
        tree = cls(course.id)
        tree.reload(course, sync_receipt=sync_receipt)
        return tree

    def __init__(
        self,
        tree_name: str,
        *,
        emit_selection_signal_for_type: Type,
    ):
        super().__init__(tree_name)
        self._reloading = False
        self._last_selected_id: Optional[str] = None
        self._target_data_type = emit_selection_signal_for_type

        self.selectionModel().selectionChanged.connect(self._selection_changed)

    def ui_setup(
        self,
        *,
        header_text: str | Sequence[str],
        indentation: int = 20,
        max_width: Optional[int] = None,
        min_width: Optional[int] = None,
        alternating_row_colours: bool = False,
    ) -> None:
        if not isinstance(header_text, str) and isinstance(header_text, Sequence):
            self.setHeaderLabels(header_text)
        else:
            self.setHeaderLabel(header_text)

        self.setIndentation(indentation)

        if max_width is not None:
            self.setMaximumWidth(max_width)

        if min_width is not None:
            self.setMinimumWidth(min_width)

        self.setAlternatingRowColors(alternating_row_colours)

    def set_columns_resize_mode(
        self,
        resize_mode_for_columns: list[QHeaderView.ResizeMode],
        *,
        stretch_last: bool = False,
    ) -> None:
        header = self.header()

        for index, mode in enumerate(resize_mode_for_columns):
            header.setSectionResizeMode(index, mode)

        header.setStretchLastSection(stretch_last)

    def reload(self, data: T, *, sync_receipt: SyncReceipt) -> None:
        self._reloading = True

        try:
            self.clear()
            self.addTopLevelItems(self.create_tree_items(data, sync_receipt))
            self.reexpand()
        finally:
            self._reloading = False

        self.reselect()

    def reselect(self) -> None:
        if self._last_selected_id is not None:
            if not self.select_ids([self._last_selected_id]):
                self._clear_selection()

    @abstractmethod
    def create_tree_items(
        self, data: T, sync_receipt: SyncReceipt
    ) -> Sequence[QTreeWidgetItem]: ...

    @Slot(QItemSelection, QItemSelection)
    def _selection_changed(self, _0: QItemSelection, _1: QItemSelection) -> None:
        if self._reloading:
            return

        if len(self.selectedItems()) > 0:
            selected = self.selectedItems()[0]
        else:
            self._clear_selection()
            return

        if self.is_unseen(selected):
            self.mark_as_seen(selected)

        if not isinstance(selected, AnyTreeDataItem):
            self._clear_selection()
            return

        data = selected.extra_data

        if isinstance(data, self._target_data_type):
            if hasattr(data, "id"):
                _logger.debug("id=%s selected", data.id)
                self._last_selected_id = data.id
                self.item_selected.emit(data)
            else:
                raise AttributeError(
                    f"Expected {self._target_data_type.__name__} to have an id attribute"
                )
        else:
            logging.warning(
                "Expected type %s, got %s instead, ignoring",
                self._target_data_type.__name__,
                type(data).__name__,
            )

            self._clear_selection()

    def _clear_selection(self) -> None:
        _logger.debug("Clearing selection")

        self._last_selected_id = None
        self.item_selected.emit(None)

    def is_unseen(self, item: QTreeWidgetItem) -> bool:
        return item.font(0).bold()

    def mark_as_unseen(self, item: QTreeWidgetItem) -> None:
        item.setFont(0, bold_font)

    def mark_as_seen(self, item: QTreeWidgetItem) -> None:
        item.setFont(0, normal_font)



import logging

from qtpy.QtCore import QObject, Signal, Slot
from qtpy.QtGui import QGuiApplication, Qt

_logger = logging.getLogger(__name__)


def colour_scheme() -> Qt.ColorScheme:
    return QGuiApplication.styleHints().colorScheme()


def is_dark_colour_scheme() -> bool:
    return colour_scheme() == Qt.ColorScheme.Dark


class ColourSchemeChangeEvent(QObject):
    theme_changed = Signal()

    def __init__(self):
        super().__init__(None)
        self._last_theme = colour_scheme()
        QGuiApplication.styleHints().colorSchemeChanged.connect(self._theme_changed)

    @Slot(Qt.ColorScheme)
    def _theme_changed(self, colour_scheme: Qt.ColorScheme) -> None:
        # Ensure the signal isn't fired when there wasn't actually a change
        if colour_scheme != self._last_theme:
            self._last_theme = colour_scheme
            self.theme_changed.emit()


_theme_changed_listener = ColourSchemeChangeEvent()


def colour_scheme_changed() -> Signal:
    global _theme_changed_listener
    return _theme_changed_listener.theme_changed



import logging
import os.path

from qtpy.QtCore import QFileInfo, QMimeDatabase
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QApplication, QFileIconProvider, QStyle

import qcanvas.util.runtime as runtime

_logger = logging.getLogger(__name__)

# Windows and linux have different ways of doing this
if runtime.is_running_on_windows:
    _icon_provider = QFileIconProvider()

    def icon_for_filename(file_name: str) -> QIcon:
        return _icon_provider.icon(QFileInfo(file_name))

else:
    _mime_database = QMimeDatabase()
    # This must be initialised lazily because the QApplication might not be initialised at this time
    _default_icon: QIcon | None = None
    _icon_for_suffix: dict[str, QIcon] = {}

    def icon_for_filename(file_name: str) -> QIcon:
        global _default_icon

        file_suffix = os.path.splitext(file_name)[1]

        # Check if we already know what icon this file type has
        if file_suffix in _icon_for_suffix:
            return _icon_for_suffix[file_suffix]

        # Try to find an icon for this file type
        for mime_type in _mime_database.mimeTypesForFileName(file_name):
            icon = QIcon.fromTheme(mime_type.iconName())

            if not icon.isNull():
                _icon_for_suffix[file_suffix] = icon
                return icon

        _lazy_init_default_icon()

        # No icon for this type of file was found, return default icon
        _icon_for_suffix[file_suffix] = _default_icon
        return _default_icon

    def _lazy_init_default_icon() -> None:
        global _default_icon

        if _default_icon is None:
            _default_icon = QApplication.style().standardIcon(
                QStyle.StandardPixmap.SP_FileIcon
            )



import logging
from typing import *

from qtpy.QtCore import Slot
from qtpy.QtGui import QAction
from qtpy.QtWidgets import QMenu

from qcanvas.util import settings

_logger = logging.getLogger(__name__)


class QuickSyncOption(QAction):
    def __init__(self, parent: Optional[QMenu] = None):
        super().__init__("Ignore old courses", parent)
        self.setToolTip(
            "When this option is selected, old courses will not be synchronised. This will only effect the first sync."
        )
        self.setCheckable(True)
        self.setChecked(settings.client.quick_sync_enabled)
        self.triggered.connect(self._triggered)

    @Slot()
    def _triggered(self) -> None:
        settings.client.quick_sync_enabled = self.isChecked()



# This file regenerates icons.qrc, rc_icons.py and the icons definition file based on the icons in dark/, light/ and universal/

import glob
import logging
import os
import subprocess
from pathlib import Path

import xmltodict

_logger = logging.getLogger(__name__)


def remove_base_directory(path: str | Path) -> Path:
    if isinstance(path, str):
        path = Path(path)

    return Path(*path.parts[1:])


def generate_xml(all_icons: list[Path]) -> str:
    xml_dict = {
        "RCC": {
            "@version": "1.0",
            "qresource": {"@prefix": "icons/", "file": all_icons},
        }
    }

    return "<!DOCTYPE RCC>\n" + xmltodict.unparse(
        xml_dict, pretty=True, full_document=False
    )


def generate_icon_defs(icons: list[Path]) -> str:
    icons_base_removed = [remove_base_directory(icon) for icon in icons]
    groups = group_icons_by_category(icons_base_removed)
    classes = [
        generate_group_class(group_name, icons) for group_name, icons in groups.items()
    ]

    return f"""from .rc_icons import qt_resource_data as _  # Without this, icon data will not be loaded
from qtpy.QtGui import QIcon
from ._icon_type import UniversalIcon, ThemeIcon

""" + "\n\n".join(
        classes
    )


def group_icons_by_category(icons: list[Path]) -> dict[str, set[Path]]:
    groups: dict[str, set[Path]] = {}

    for icon in icons:
        icon_category = icon.parts[0]
        groups.setdefault(icon_category, set())
        groups[icon_category].add(icon)

    return groups


def generate_group_class(group_name: str, icons: set[Path]) -> str:
    # pascal_case_class_name = "".join([part.title() for part in group_name.split("_")])
    lines = ["# noinspection PyPep8Naming", f"class {group_name}:"]

    for icon in sorted(icons):
        path_no_ext = os.path.splitext(icon)[0]
        lines.append(f'    {icon.stem} = QIcon.fromTheme("{path_no_ext}")')

    return "\n".join(lines)


def run_rcc() -> None:
    subprocess.run(["pyside6-rcc", "icons.qrc", "-o", "rc_icons.py"])


def write(file_name: str, content: str) -> None:
    with open(file_name, "w") as f:
        f.write(content)


if __name__ == "__main__":
    all_icons: list[Path] = [
        Path(file) for file in glob.glob("**/*.svg", recursive=True)
    ]
    all_icons.sort()

    write("icons.qrc", generate_xml(all_icons))
    write("__init__.py", generate_icon_defs(all_icons))
    run_rcc()



import logging

from qtpy.QtCore import QByteArray, QSettings

from qcanvas.util.settings._mapped_setting import MappedSetting
from qcanvas.util.themes import default_theme, ensure_theme_is_valid

_logger = logging.getLogger(__name__)


class ThemeSetting(MappedSetting):
    def __init__(self):
        super().__init__(default=default_theme)

    def __get__(self, instance, owner):
        return ensure_theme_is_valid(super().__get__(instance, owner))

    def __set__(self, instance, value):
        super().__set__(instance, ensure_theme_is_valid(value))


class _UISettings:
    settings = QSettings("QCanvasTeam", "QCanvas")
    theme: ThemeSetting = ThemeSetting()
    last_geometry: MappedSetting[QByteArray] = MappedSetting()
    last_window_state: MappedSetting[QByteArray] = MappedSetting()



import os
import platform
import sys
from pathlib import Path

is_running_portable = Path(".portable").exists()
is_running_on_windows = platform.system() == "Windows"
is_running_on_linux = platform.system() == "Linux"
_is_nuitka = "__compiled__" in globals()
_is_pyinstaller = getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")
is_running_as_compiled = _is_nuitka or _is_pyinstaller
is_running_as_flatpak = os.environ.get("container", "") == "flatpak"

__all__ = [
    "is_running_portable",
    "is_running_as_flatpak",
    "is_running_on_linux",
    "is_running_as_compiled",
    "is_running_on_windows",
]



import logging
from dataclasses import dataclass

import qcanvas_backend.database.types as db
from qcanvas_backend.net.resources.download.resource_manager import ResourceManager
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qtpy.QtCore import Slot
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import *

from qcanvas import icons
from qcanvas.ui.course_viewer.tabs.assignment_tab import AssignmentTab
from qcanvas.ui.course_viewer.tabs.mail_tab import MailTab
from qcanvas.ui.course_viewer.tabs.page_tab import PageTab
from qcanvas.util.basic_fonts import bold_font
from qcanvas.util.layouts import layout
from qcanvas.util.ui_tools import make_truncatable

_logger = logging.getLogger(__name__)


@dataclass
class _Tab:
    icon: QIcon
    highlighted_icon: QIcon


class CourseViewer(QWidget):

    def __init__(
        self,
        course: db.Course,
        downloader: ResourceManager,
        *,
        sync_receipt: SyncReceipt,
    ):
        super().__init__()
        # todo this is a mess. there are several other messes like this too, do they all have to be a mess?
        self._course_id = course.id
        self._previous_tab_index = 0

        self._course_label = QLabel(course.name)
        self._course_label.setFont(bold_font)
        make_truncatable(self._course_label)

        self._pages_tab = PageTab.create_from_receipt(
            course=course,
            downloader=downloader,
            sync_receipt=sync_receipt,
        )
        self._assignments_tab = AssignmentTab.create_from_receipt(
            course=course,
            downloader=downloader,
            sync_receipt=sync_receipt,
        )
        self._mail_tab = MailTab.create_from_receipt(
            course=course,
            downloader=downloader,
            sync_receipt=sync_receipt,
        )
        # self._files_tab = FileTab.create_from_receipt(
        #     course=course,
        #     downloader=downloader,
        #     sync_receipt=sync_receipt,
        # )

        self._tab_widget = QTabWidget()
        self._tabs: dict[int, _Tab] = {}

        # self._setup_tab(
        #     name="Files",
        #     widget=self._files_tab,
        #     icon=icons.tabs.pages,
        #     highlighted_icon=icons.tabs.pages_new_content,
        # )
        self._PAGES_TAB = self._setup_tab(
            name="Pages",
            widget=self._pages_tab,
            icon=icons.tabs.pages,
            highlighted_icon=icons.tabs.pages_new_content,
        )
        self._ASSIGNMENTS_TAB = self._setup_tab(
            name="Assignments",
            widget=self._assignments_tab,
            icon=icons.tabs.assignments,
            highlighted_icon=icons.tabs.assignments_new_content,
        )
        self._MAIL_TAB = self._setup_tab(
            name="Mail",
            widget=self._mail_tab,
            icon=icons.tabs.mail,
            highlighted_icon=icons.tabs.mail_new_content,
        )
        # self._tabs.addTab(QLabel("Not implemented"), "Panopto")  # The meme lives on!

        self.setLayout(layout(QVBoxLayout, self._course_label, self._tab_widget))
        self._tab_widget.currentChanged.connect(self._tab_changed)
        self._highlight_tabs(sync_receipt)

    def _setup_tab(
        self, widget: QWidget, icon: QIcon, highlighted_icon: QIcon, name: str
    ) -> int:
        index = self._tab_widget.addTab(widget, icon, name)
        self._tabs[index] = _Tab(icon, highlighted_icon)
        return index

    def reload(self, course: db.Course, *, sync_receipt: SyncReceipt) -> None:
        # self._files_tab.reload(course, sync_receipt=sync_receipt)
        self._pages_tab.reload(course, sync_receipt=sync_receipt)
        self._assignments_tab.reload(course, sync_receipt=sync_receipt)
        self._mail_tab.reload(course, sync_receipt=sync_receipt)
        self._highlight_tabs(sync_receipt)

    @Slot(int)
    def _tab_changed(self, index: int) -> None:
        _logger.debug(f"Index = {index}")
        if index != -1:
            _logger.debug(f"Previous tab = {self._previous_tab_index}")
            self._unhighlight_tab(self._previous_tab_index)
            self._previous_tab_index = index

    def _highlight_tabs(self, sync_receipt: SyncReceipt) -> None:
        updates = sync_receipt.updates_by_course.get(self._course_id, None)

        if updates is not None:
            if len(updates.updated_pages) > 0:
                self._highlight_tab(self._PAGES_TAB)
            else:
                self._unhighlight_tab(self._PAGES_TAB)

            if len(updates.updated_assignments) > 0:
                self._highlight_tab(self._ASSIGNMENTS_TAB)
            else:
                self._unhighlight_tab(self._ASSIGNMENTS_TAB)

            if len(updates.updated_messages) > 0:
                self._highlight_tab(self._MAIL_TAB)
            else:
                self._unhighlight_tab(self._MAIL_TAB)
        else:
            for index in range(0, len(self._tabs)):
                self._unhighlight_tab(index)

    def _highlight_tab(self, tab_index: int) -> None:
        tab = self._tabs[tab_index]
        self._tab_widget.setTabIcon(tab_index, tab.highlighted_icon)

    def _unhighlight_tab(self, tab_index: int) -> None:
        tab = self._tabs[tab_index]
        self._tab_widget.setTabIcon(tab_index, tab.icon)



import logging
from typing import *

from qtpy.QtWidgets import QTreeWidgetItem

_logger = logging.getLogger(__name__)


class MemoryTreeWidgetItem(QTreeWidgetItem):
    def __init__(
        self, id: str, data: Optional[object], strings: Optional[List[str]] = None
    ):
        super().__init__(strings)
        self._id = id
        self.extra_data = data

    @property
    def id(self) -> str:
        return self._id



import logging
from typing import *

import qcanvas_backend.database.types as db
from qcanvas_backend.net.resources.download.resource_manager import ResourceManager
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qtpy.QtWidgets import *

from qcanvas.ui.course_viewer.tabs.content_tab import ContentTab
from qcanvas.ui.course_viewer.tabs.mail_tab.mail_tree import MailTree
from qcanvas.ui.course_viewer.tabs.util import date_strftime_format
from qcanvas.util.basic_fonts import bold_label
from qcanvas.util.layouts import grid_layout

_logger = logging.getLogger(__name__)


# todo maybe update has_been_read? probably not the responsibility of this class though
class MailTab(ContentTab):
    def __init__(
        self,
        *,
        course: db.Course,
        sync_receipt: SyncReceipt,
        downloader: ResourceManager,
    ):
        super().__init__(
            explorer=MailTree.create_from_receipt(course, sync_receipt=sync_receipt),
            title_placeholder_text="No mail selected",
            downloader=downloader,
        )

        self._date_sent_label = QLabel("")
        self._sender_label = QLabel("")

        self.enable_info_grid()

    def setup_info_grid(self) -> QGridLayout:
        grid = grid_layout(
            [
                [
                    bold_label("From:"),
                    self._sender_label,
                ],
                [
                    bold_label("Date:"),
                    self._date_sent_label,
                ],
            ]
        )

        grid.setColumnStretch(0, 0)
        grid.setColumnStretch(1, 1)

        return grid

    def update_info_grid(self, mail: db.CourseMessage) -> None:
        self._date_sent_label.setText(mail.creation_date.strftime(date_strftime_format))
        self._sender_label.setText(mail.sender_name)



import logging
from typing import *

import qcanvas_backend.database.types as db
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qtpy.QtCore import Qt, Signal

from qcanvas import icons
from qcanvas.ui.course_viewer.content_tree import ContentTree
from qcanvas.ui.course_viewer.course_tree._course_icon_generator import (
    CourseIconGenerator,
)
from qcanvas.ui.course_viewer.tree_widget_data_item import TreeWidgetDataItem
from qcanvas.ui.memory_tree import MemoryTreeWidgetItem

_logger = logging.getLogger(__name__)


class _CourseTreeItem(TreeWidgetDataItem):
    def __init__(self, course: db.Course, owner: "CourseTree"):
        TreeWidgetDataItem.__init__(
            self,
            id=course.id,
            data=course,
            strings=[course.configuration.nickname or course.name],
        )

        self._owner = owner
        self._course = course

        self.setFlags(
            Qt.ItemFlag.ItemIsEditable
            | Qt.ItemFlag.ItemIsSelectable
            | Qt.ItemFlag.ItemIsEnabled
        )

    def setData(self, column: int, role: int, value: Any):
        if column != 0 or not isinstance(value, str):
            return super().setData(column, role, value)

        value = value.strip()

        if len(value) == 0:
            super().setData(column, role, self._course.name)
            self._owner.course_renamed.emit(self._course, None)
        else:
            super().setData(column, role, value)
            self._owner.course_renamed.emit(self._course, value)


class CourseTree(ContentTree[Sequence[db.Term]]):
    course_renamed = Signal(db.Course, str)

    def __init__(self):
        super().__init__("course_tree", emit_selection_signal_for_type=db.Course)

        self.ui_setup(
            max_width=250, min_width=150, header_text="Courses", indentation=15
        )

    def create_tree_items(
        self, terms: List[db.Term], sync_receipt: SyncReceipt
    ) -> Sequence[MemoryTreeWidgetItem]:
        widgets = []

        for term in reversed(terms):
            term_widget = self._create_term_widget(term)
            course_icon_generator = CourseIconGenerator(term.id)

            for course in term.courses:
                course_widget = self._create_course_widget(
                    course, course_icon_generator, sync_receipt
                )
                term_widget.addChild(course_widget)

            widgets.append(term_widget)

        return widgets

    def _create_course_widget(
        self,
        course: db.Course,
        course_icon_generator: CourseIconGenerator,
        sync_receipt: SyncReceipt,
    ) -> _CourseTreeItem:
        course_widget = _CourseTreeItem(course, self)
        course_widget.setIcon(0, course_icon_generator.get_icon())

        if sync_receipt.was_updated(course):
            self.mark_as_unseen(course_widget)

        return course_widget

    def _create_term_widget(self, term: db.Term) -> MemoryTreeWidgetItem:
        term_widget = MemoryTreeWidgetItem(id=term.id, data=term, strings=[term.name])
        term_widget.setFlags(Qt.ItemFlag.ItemIsEnabled)
        term_widget.setIcon(0, icons.tree_items.semester)

        return term_widget



from qtpy.QtGui import QFont
from qtpy.QtWidgets import QLabel

normal_font = QFont()
bold_font = QFont()
bold_font.setBold(True)


def bold_label(text: str) -> QLabel:
    result = QLabel(text)
    result.setFont(bold_font)
    return result



import logging
from typing import *

from qtpy.QtCore import Slot
from qtpy.QtGui import QAction
from qtpy.QtWidgets import QMenu

from qcanvas.util import settings

_logger = logging.getLogger(__name__)


class SyncOnStartOption(QAction):
    def __init__(self, parent: Optional[QMenu] = None):
        super().__init__("Sync on start", parent)
        self.setToolTip(
            "When this option is selected, synchronisation will be started automatically when the app starts."
        )
        self.setCheckable(True)
        self.setChecked(settings.client.sync_on_start)
        self.triggered.connect(self._triggered)

    @Slot()
    def _triggered(self) -> None:
        settings.client.sync_on_start = self.isChecked()



import html
import logging
from typing import Optional

import qcanvas_backend.database.types as db
from bs4 import BeautifulSoup, Tag
from qasync import asyncSlot
from qcanvas_backend.net.resources.download.resource_manager import ResourceManager
from qcanvas_backend.net.resources.extracting.no_extractor_error import NoExtractorError
from qcanvas_backend.util import is_link_invisible
from qtpy.QtCore import QUrl, Slot
from qtpy.QtGui import QDesktopServices
from qtpy.QtWidgets import QTextBrowser

from qcanvas.backend_connectors import FrontendResourceManager
from qcanvas.util.html_cleaner import clean_up_html
from qcanvas.util.qurl_util import file_url

_logger = logging.getLogger(__name__)


class ResourceRichBrowser(QTextBrowser):
    def __init__(self, downloader: ResourceManager):
        super().__init__()
        self._downloader = downloader
        self._content: Optional[db.CourseContentItem] = None
        self._current_content_resources: dict[str, db.Resource] = {}
        self._extractors = downloader.extractors
        self.setMinimumWidth(300)
        self.setOpenLinks(False)
        self.anchorClicked.connect(self._open_url)

        if isinstance(self._downloader, FrontendResourceManager):
            self._downloader.download_finished.connect(self._download_updated)
            self._downloader.download_failed.connect(self._download_updated)

    def show_blank(self, completely_blank: bool = False) -> None:
        if completely_blank:
            self.clear()
        else:
            self.setPlainText("No content")

        self._content = None
        self._current_content_resources.clear()

    def show_content(self, page: db.CourseContentItem) -> None:
        if page.body is None:
            self.show_blank()
        else:
            self._collect_resources(page)
            self._show_page_content(page)

    def _collect_resources(self, page: db.CourseContentItem):
        self._current_content_resources = {
            resource.id: resource for resource in page.resources
        }

    def _show_page_content(self, page: db.CourseContentItem):
        self._content = page
        html = clean_up_html(page.body)
        html = self._substitute_links(html)
        self.setHtml(html)

    def _substitute_links(self, html: str) -> str:
        doc = BeautifulSoup(html, "html.parser")

        for resource_link in doc.find_all(self._extractors.tag_whitelist):
            try:
                extractor = self._extractors.extractor_for_tag(resource_link)
                resource_id = extractor.resource_id_from_tag(resource_link)

                if is_link_invisible(resource_link):
                    _logger.debug("Found dead link for %s, removing", resource_id)
                    resource_link.decompose()
                    continue
                elif resource_id not in self._current_content_resources:
                    _logger.debug(
                        "%s not found in page resources, ignoring", resource_id
                    )
                    continue

                file_link_tag = self._create_resource_link_tag(
                    resource_id, resource_link.name == "img"
                )
                resource_link.replace_with(file_link_tag)
            except NoExtractorError:
                pass

        return str(doc)

    def _create_resource_link_tag(self, resource_id: str, is_image: bool) -> Tag:
        resource = self._current_content_resources[resource_id]

        # todo not sure if this is a good idea or not
        # if is_image and resource.download_state == db.ResourceDownloadState.DOWNLOADED:
        #     location = self._downloader.resource_download_location(resource)
        #
        #     file_link_tag = doc.new_tag(
        #         "img",
        #         attrs={
        #             "source": location.absolute(),
        #         },
        #     )
        # else:

        return BeautifulSoup(
            markup=f"""
            <a href="data:{html.escape(resource_id)}" style="font-weight: normal;">
                <img height="18" src="{html.escape(self._download_state_icon(resource.download_state))}"/>
                {html.escape(resource.file_name)}
            </a>
            """,
            features="html.parser",
        ).a

    def _download_state_icon(self, download_state: db.ResourceDownloadState) -> str:
        icon_path = ":icons/universal/downloads"
        match download_state:
            case db.ResourceDownloadState.DOWNLOADED:
                return f"{icon_path}/downloaded.svg"
            case db.ResourceDownloadState.NOT_DOWNLOADED:
                return f"{icon_path}/not_downloaded.svg"
            case db.ResourceDownloadState.FAILED:
                return f"{icon_path}/download_failed.svg"
            case _:
                raise ValueError()

    @asyncSlot(QUrl)
    async def _open_url(self, url: QUrl) -> None:
        if url.scheme() == "data":
            await self._open_resource_from_link(url)
        else:
            QDesktopServices.openUrl(url)

    async def _open_resource_from_link(self, url) -> None:
        resource_id = url.path()
        resource = self._current_content_resources[resource_id]

        try:
            await self._downloader.download(resource)
        except Exception as e:
            _logger.warning(
                "Download of resource id=%s failed", resource_id, exc_info=e
            )
            return

        resource_path = file_url(self._downloader.resource_download_location(resource))
        QDesktopServices.openUrl(resource_path)

    @Slot(db.Resource)
    def _download_updated(self, resource: db.Resource) -> None:
        if self._content is not None and resource.id in self._current_content_resources:
            # BANDAID FIX: In the following situation:
            # - Download is started
            # - Synchronisation is started
            # - Download finishes AFTER the sync
            # --> `resource` is NOT `self._current_content_resources[resource.id]`, because the sync will reload the
            # resource from the DB, but the downloader will still only know about the old resource object.
            # This causes resources not update their download state in the viewer. This line "fixes" that, but does NOT
            # address the root cause. I think reloading the resource from the DB somewhere is the only true fix for this

            if self._current_content_resources[resource.id] is not resource:
                _logger.warning(
                    "Resource has diverged from current loaded data, applying bandaid fix"
                )
                self._current_content_resources[resource.id].download_state = (
                    resource.download_state
                )

            self._show_page_content(self._content)



import logging

from qtpy.QtCore import QObject, Signal

_logger = logging.getLogger(__name__)


class ThemeChangedEvent(QObject):
    theme_changed = Signal()


theme_changed_event = ThemeChangedEvent(None)


def theme_changed() -> Signal:
    global theme_changed_event
    return theme_changed_event.theme_changed



import logging
from typing import *

from qtpy.QtCore import QItemSelectionModel, Slot
from qtpy.QtWidgets import QTreeWidget, QTreeWidgetItem, QWidget

from qcanvas.ui.memory_tree._tree_memory import TreeMemory
from qcanvas.ui.memory_tree.memory_tree_widget_item import MemoryTreeWidgetItem

_logger = logging.getLogger(__name__)


class MemoryTreeWidget(QTreeWidget):
    def __init__(
        self,
        tree_name: str,
        parent: Optional[QWidget] = None,
    ):
        super().__init__(parent)
        self._id_map: dict[str, QTreeWidgetItem] = {}
        self._memory = TreeMemory(tree_name)
        self._suppress_expansion_signals = False
        self._suppress_selection_signal = False

        self.itemExpanded.connect(self._expanded)
        self.itemCollapsed.connect(self._collapsed)

    def reexpand(self) -> None:
        self.scheduleDelayedItemsLayout()

        try:
            self._suppress_expansion_signals = True

            collapsed_ids = self._memory.collapsed_ids

            for widget in self._id_map.values():
                if widget.id not in collapsed_ids:
                    _logger.debug("Re-expand %s", widget.id)
                    self.expand(self.indexFromItem(widget, 0))
        finally:
            self._suppress_expansion_signals = False

    def clear(self):
        super().clear()
        self._id_map.clear()

    def select_ids(self, ids: List[str]) -> bool:
        """
        :returns: True if all ids were still found in the tree, False if one or more was missing
        """
        self._suppress_selection_signal = True

        is_first = True
        all_ids_in_tree = True

        try:
            for widget_id in ids:
                if widget_id in self._id_map:
                    _logger.debug("Selected %s", widget_id)

                    flags = (
                        QItemSelectionModel.SelectionFlag.Rows
                        | QItemSelectionModel.SelectionFlag.Select
                    )

                    if is_first:
                        flags |= QItemSelectionModel.SelectionFlag.Clear

                    self.selectionModel().select(
                        self.indexFromItem(self._id_map[widget_id], 0), flags
                    )
                else:
                    _logger.debug(
                        "Item %s is no longer in the tree, can't select it", widget_id
                    )
                    all_ids_in_tree = False
        finally:
            self._suppress_selection_signal = False

        return all_ids_in_tree

    def insertTopLevelItem(self, index: int, item: QTreeWidgetItem):
        super().insertTopLevelItem(index, item)
        self._add_widget_to_id_map(item)

    def insertTopLevelItems(self, index: int, items: Sequence[QTreeWidgetItem]):
        super().insertTopLevelItems(index, items)
        self._add_widget_to_id_map(items)

    def addTopLevelItems(self, items: Sequence[QTreeWidgetItem]):
        super().addTopLevelItems(items)
        self._add_widget_to_id_map(items)

    def _add_widget_to_id_map(
        self, widget: QTreeWidgetItem | Sequence[QTreeWidgetItem]
    ):
        map_updates = {}
        widget_stack = widget if isinstance(widget, List) else [widget]

        while len(widget_stack) > 0:
            item = widget_stack.pop()

            if hasattr(item, "id"):
                if item.id in self._id_map or item.id in map_updates:
                    raise ValueError(f"Item with ID {item.id} is already in the tree")

                map_updates[item.id] = item
                _logger.debug("Add %s to map", item.id)

            if item.childCount() > 0:
                widget_stack.extend(
                    [item.child(index) for index in range(0, item.childCount())]
                )

        self._id_map.update(map_updates.items())

    @Slot(QTreeWidgetItem)
    def _expanded(self, item: QTreeWidgetItem):
        if self._suppress_expansion_signals:
            return

        if isinstance(item, MemoryTreeWidgetItem):
            _logger.debug("Expanded %s", item.id)
            self._memory.expanded(item.id)

    @Slot(QTreeWidgetItem)
    def _collapsed(self, item: QTreeWidgetItem):
        if self._suppress_expansion_signals:
            return

        if isinstance(item, MemoryTreeWidgetItem):
            _logger.debug("Collapsed %s", item.id)
            self._memory.collapsed(item.id)



import logging
from typing import Any

from qtpy.QtGui import QIcon, QKeySequence, QPixmap
from qtpy.QtWidgets import *

_logger = logging.getLogger(__name__)


def make_truncatable(widget: QWidget) -> None:
    widget.setSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Fixed)


def create_qaction(
    *,
    name: str,
    shortcut: QKeySequence | None = None,
    parent: QMenu = None,
    triggered: Any = None,
    checkable: bool | None = None,
    checked: bool | None = None,
    icon: QIcon | QPixmap | None = None
) -> QAction:
    action = QAction(name)

    if shortcut is not None:
        action.setShortcut(shortcut)

    if parent is not None:
        action.setParent(parent)
        parent.addAction(action)

    if triggered is not None:
        action.triggered.connect(triggered)

    if checkable is not None:
        action.setCheckable(checkable)

        if checked is not None:
            action.setChecked(checked)

    if icon is not None:
        action.setIcon(icon)

    return action



import logging

import qcanvas_backend.database.types as db
from qcanvas_backend.net.resources.download.resource_manager import ResourceManager
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt
from qtpy.QtWidgets import *

from qcanvas.ui.course_viewer.tabs.file_tab.pages_file_tree import PagesFileTree
from qcanvas.util.layouts import layout

_logger = logging.getLogger(__name__)


class FileTab(QWidget):
    @staticmethod
    def create_from_receipt(
        *,
        course: db.Course,
        sync_receipt: SyncReceipt,
        downloader: ResourceManager,
    ) -> "FileTab":
        return FileTab(course=course, sync_receipt=sync_receipt, downloader=downloader)

    def __init__(
        self,
        course: db.Course,
        downloader: ResourceManager,
        *,
        sync_receipt: SyncReceipt = None,
    ):
        super().__init__()

        self._page_file_tree = PagesFileTree.create_from_receipt(
            course, sync_receipt=sync_receipt, resource_manager=downloader
        )
        # self._assignment_file_tree = FileTree.create_from_receipt(
        #     course,
        #     sync_receipt=sync_receipt,
        #     mode="assignments",
        #     resource_manager=downloader,
        # )

        self.setLayout(layout(QHBoxLayout, self._page_file_tree))

    def reload(self, course: db.Course, *, sync_receipt: SyncReceipt) -> None:
        pass



from qtpy.QtGui import QIcon

from ._icon_type import ThemeIcon, UniversalIcon
from .rc_icons import (
    qt_resource_data as _,  # Without this, icon data will not be loaded
)


# noinspection PyPep8Naming
class actions:
    exit = QIcon.fromTheme("actions/exit")
    mark_all_read = QIcon.fromTheme("actions/mark_all_read")
    open_downloads = QIcon.fromTheme("actions/open_downloads")
    quick_login = QIcon.fromTheme("actions/quick_login")
    sync = QIcon.fromTheme("actions/sync")


# noinspection PyPep8Naming
class branding:
    logo_transparent = QIcon.fromTheme("branding/logo_transparent")
    main_icon = QIcon.fromTheme("branding/main_icon")


# noinspection PyPep8Naming
class options:
    auto_download = QIcon.fromTheme("options/auto_download")
    ignore_old = QIcon.fromTheme("options/ignore_old")
    include_videos = QIcon.fromTheme("options/include_videos")
    theme = QIcon.fromTheme("options/theme")


# noinspection PyPep8Naming
class tabs:
    assignments = QIcon.fromTheme("tabs/assignments")
    assignments_new_content = QIcon.fromTheme("tabs/assignments_new_content")
    mail = QIcon.fromTheme("tabs/mail")
    mail_new_content = QIcon.fromTheme("tabs/mail_new_content")
    pages = QIcon.fromTheme("tabs/pages")
    pages_new_content = QIcon.fromTheme("tabs/pages_new_content")


# noinspection PyPep8Naming
class tree_items:
    assignment = QIcon.fromTheme("tree_items/assignment")
    mail = QIcon.fromTheme("tree_items/mail")
    module = QIcon.fromTheme("tree_items/module")
    page = QIcon.fromTheme("tree_items/page")
    semester = QIcon.fromTheme("tree_items/semester")


# noinspection PyPep8Naming
class downloads:
    download_failed = QIcon.fromTheme("downloads/download_failed")
    downloaded = QIcon.fromTheme("downloads/downloaded")
    not_downloaded = QIcon.fromTheme("downloads/not_downloaded")
    unknown = QIcon.fromTheme("downloads/unknown")



import logging

from qcanvas.util.settings._client_settings import _ClientSettings
from qcanvas.util.settings._ui_settings import _UISettings

_logger = logging.getLogger(__name__)

client = _ClientSettings()
ui = _UISettings()



import logging
from pathlib import Path
from typing import *

from lightdb import LightDB, Model

from qcanvas.util import paths

_logger = logging.getLogger(__name__)


def _storage_path() -> Path:
    path = paths.ui_storage() / "TREE.DB"
    path.parent.mkdir(parents=True, exist_ok=True)
    return path


_state_db = LightDB(str(_storage_path()))


class _TreeState(Model, table="trees", db=_state_db):
    tree_name: str
    collapsed_items: List[str] = []


def _get_or_create_state(name: str) -> _TreeState:
    state = _TreeState.get(tree_name=name)

    if state is None:
        state = _TreeState.create(tree_name=name)
        # Initialise the list here! Or else every instance has the same list object
        state.collapsed_items = []
        # Important or instances will get duplicated data in some cases
        state.save()
        return state
    else:
        return state


class TreeMemory:
    def __init__(self, tree_name: str):
        self._tree_name = tree_name
        self._state = _get_or_create_state(tree_name)

    def is_expanded(self, node_id: str) -> bool:
        return node_id in self._state.expanded_items

    def expanded(self, node_id: str) -> None:
        self.set_expanded(node_id, True)

    def collapsed(self, node_id: str) -> None:
        self.set_expanded(node_id, False)

    def set_expanded(self, node_id: str, expanded: bool) -> None:
        contains = node_id in self._state.collapsed_items

        # fixme when using a slow usb stick, this can momentarily block the event loop.
        if expanded and contains:
            self._state.collapsed_items.remove(node_id)
            self._state.save()
        elif not expanded and not contains:
            self._state.collapsed_items.append(node_id)
            self._state.save()

    @property
    def collapsed_ids(self) -> List[str]:
        return self._state.collapsed_items



import logging

import qdarktheme
from qtpy.QtCore import Slot
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QApplication, QStyleFactory

from qcanvas.util.themes._colour_scheme_helper import (
    colour_scheme_changed,
    is_dark_colour_scheme,
)
from qcanvas.util.themes._selected_theme import SelectedTheme
from qcanvas.util.themes._theme_changed_event import theme_changed

_logger = logging.getLogger(__name__)

default_theme = "auto"
_is_dark_mode: bool | None = None
_selected_theme: SelectedTheme | None = None

_universal_path = ":icons/universal"
_dark_path = ":icons/dark"
_light_path = ":icons/light"


def apply(theme: str) -> None:
    global _is_dark_mode, _selected_theme

    theme = ensure_theme_is_valid(theme)
    was_dark_mode = _is_dark_mode

    if theme != "native":
        if theme == "auto":
            _is_dark_mode = is_dark_colour_scheme()
            _selected_theme = SelectedTheme.AUTO
            selected_colour_scheme = "dark" if _is_dark_mode else "light"
        else:
            _is_dark_mode = theme == "dark"
            _selected_theme = SelectedTheme.OVERRIDE
            selected_colour_scheme = theme

        qdarktheme.setup_theme(
            selected_colour_scheme,
            custom_colors={"primary": "e02424"},
        )

        QApplication.setStyle(QStyleFactory.create("Fusion"))
    else:
        _is_dark_mode = is_dark_colour_scheme()
        _selected_theme = SelectedTheme.NATIVE

    if was_dark_mode != _is_dark_mode:
        _set_fallback_paths()
        theme_changed().emit()


def is_dark_mode() -> bool:
    return _is_dark_mode


def ensure_theme_is_valid(theme: str) -> str:
    if theme not in ["auto", "light", "dark", "native"]:
        return default_theme
    else:
        return theme


@Slot()
def _scheme_changed():
    global _selected_theme, _is_dark_mode

    if _selected_theme == SelectedTheme.AUTO:
        apply("auto")
    elif _selected_theme == SelectedTheme.NATIVE:
        _is_dark_mode = is_dark_colour_scheme()
        _set_fallback_paths()
        theme_changed().emit()


def _set_fallback_paths():
    QIcon.setFallbackSearchPaths(
        [_dark_path if _is_dark_mode else _light_path, _universal_path]
    )


colour_scheme_changed().connect(_scheme_changed)



import logging

import qcanvas_backend.database.types as db
from qcanvas_backend.net.resources.download.resource_manager import ResourceManager
from qcanvas_backend.net.sync.sync_receipt import SyncReceipt

from qcanvas.ui.course_viewer.tabs.content_tab import ContentTab
from qcanvas.ui.course_viewer.tabs.page_tab.page_tree import PageTree

_logger = logging.getLogger(__name__)


class PageTab(ContentTab):
    def __init__(
        self,
        *,
        course: db.Course,
        sync_receipt: SyncReceipt,
        downloader: ResourceManager,
    ):
        super().__init__(
            explorer=PageTree.create_from_receipt(course, sync_receipt=sync_receipt),
            title_placeholder_text="No page selected",
            downloader=downloader,
        )



import itertools
import logging
import random

from cachetools import cached
from qtpy.QtCore import QByteArray
from qtpy.QtGui import QColor, QPainter, QPixmap
from qtpy.QtSvg import QSvgRenderer

_logger = logging.getLogger(__name__)
_transparent = QColor("#00000000")
_colours = [
    QColor(f"#{colour}")
    for colour in [
        "2ad6cb",
        "2d50ed",
        "7a10e4",
        "c61aaf",
        "d91b1b",
        "c7541b",
        "facd07",  # facd07
        "a9cf12",
    ]
]


class CourseIconGenerator:
    @staticmethod
    @cached(cache={})
    def get_for_term(term_id: str) -> "CourseIconGenerator":
        return CourseIconGenerator(term_id)

    def __init__(self, term_id: str):
        shuffled = list(_colours)

        # This is the dumbest way I've ever seen a language implement setting a seed for a RNG.
        # WTF python?! Why???
        random.seed(term_id)
        random.shuffle(shuffled)

        self._iterator = itertools.cycle(shuffled)

    def get_icon(self) -> QPixmap:
        return _create_icon(self._get_colour())

    def _get_colour(self) -> QColor:
        return next(self._iterator)


@cached(cache={}, key=lambda colour: colour.name(QColor.NameFormat.HexRgb))
def _create_icon(base_colour: QColor) -> QPixmap:
    dark_colour = QColor.fromHslF(
        base_colour.hslHueF(),
        base_colour.hslSaturationF(),
        base_colour.lightnessF() * 0.6875,
    )

    result_pixmap = QPixmap(256, 256)
    result_pixmap.fill(_transparent)

    with (painter := QPainter(result_pixmap)):
        svg = _create_svg_from_colours(base_colour, dark_colour)
        svg.render(painter)

    return result_pixmap


def _create_svg_from_colours(light_colour: QColor, dark_colour: QColor) -> QSvgRenderer:
    # Original SVG is from SVGRepo.com
    return QSvgRenderer(
        QByteArray(
            f"""
            <?xml version="1.0" encoding="UTF-8" standalone="no"?>
            <svg width="800px" height="800px" viewBox="0 0 24 24" >
             <g transform="translate(0 -1028.4)">
              <path d="m3 8v2 1 3 1 5 1c0 1.105 0.8954 2 2 2h14c1.105 0 2-0.895 2-2v-1-5-4-3h-18z" transform="translate(0 1028.4)" fill="{dark_colour.name()}"/>
              <path d="m3 1035.4v2 1 3 1 5 1c0 1.1 0.8954 2 2 2h14c1.105 0 2-0.9 2-2v-1-5-4-3h-18z" fill="#ecf0f1"/>
              <path d="m3 1034.4v2 1 3 1 5 1c0 1.1 0.8954 2 2 2h14c1.105 0 2-0.9 2-2v-1-5-4-3h-18z" fill="#bdc3c7"/>
              <path d="m3 1033.4v2 1 3 1 5 1c0 1.1 0.8954 2 2 2h14c1.105 0 2-0.9 2-2v-1-5-4-3h-18z" fill="#ecf0f1"/>
              <path d="m5 1c-1.1046 0-2 0.8954-2 2v1 4 2 1 3 1 5 1c0 1.105 0.8954 2 2 2h2v-1h-1.5c-0.8284 0-1.5-0.672-1.5-1.5s0.6716-1.5 1.5-1.5h12.5 1c1.105 0 2-0.895 2-2v-1-5-4-3-1c0-1.1046-0.895-2-2-2h-4-10z" transform="translate(0 1028.4)" fill="{dark_colour.name()}"/>
              <path d="m8 1v18h1 9 1c1.105 0 2-0.895 2-2v-1-5-4-3-1c0-1.1046-0.895-2-2-2h-4-6-1z" transform="translate(0 1028.4)" fill="{light_colour.name()}"/>
             </g>
            </svg>
            """.strip().encode()
        )
    )
